#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "slc.hh"
#include "options.hh"
#include "connection.hh"
#include "screen.hh"
#include "error.hh"

pthread_mutex_t  mtx_socket = PTHREAD_MUTEX_INITIALIZER;

Connection::Connection() : socket_(0) {}

bool Connection::connected() const {
  return socket_ > 0;
}

void Connection::start() {
  if (connected())
    return;

  // make connection
  connect_(O.server.c_str(), O.port);

  // send login info

  // send pass info

  // Start reader thread
  pthread_create(&g_thread_reader, 0, thread_reader, 0);
}

void Connection::stop() {
  if (!connected())
    return;
  disconnect_();
}

void Connection::connect_(const char *addr, int port) {
  struct sockaddr_in daddr;
  struct hostent *h;

  // Check no existing connexion
  assert(socket_ == 0);

  // retrieve remote host info
  h = gethostbyname(addr);
  if (h == 0) {
    S << "Unable to resolve: " << addr  << ": " << hstrerror(h_errno) << ".\n";
    return;
  }

  // create socket
  pthread_mutex_lock(&mtx_socket);
  socket_ = socket(PF_INET, SOCK_STREAM, 0);
  if (socket_ == -1) {
    S << "Unable to create socket: " << strerror(errno) << ".\n";
    socket_ = 0;
    pthread_mutex_unlock(&mtx_socket);
    return;
  }
  pthread_mutex_unlock(&mtx_socket);

  daddr.sin_family = AF_INET;
  daddr.sin_port = htons(port);
  daddr.sin_addr = *((struct in_addr *) h->h_addr);
  memset(daddr.sin_zero, '\0', sizeof daddr.sin_zero);

  S << "Connecting to "  << h->h_name  << " (" << inet_ntoa(*(struct in_addr*)h->h_addr)
    << ") on port " << port << "...\n";

  // connect
  pthread_mutex_lock(&mtx_socket);
  int con = connect(socket_, (struct sockaddr *) &daddr, sizeof daddr);
  pthread_mutex_unlock(&mtx_socket);
  if (con == -1) {
    disconnect_();
    S << "Unable to connect: " << strerror(errno);
    return;
  }

  S << "Connected to " << h->h_name << " (" << inet_ntoa(*(struct in_addr*)h->h_addr)
    << ") on port " << port << "...\n";
}

void Connection::disconnect_() {
  assert(socket_ > 0);

  pthread_mutex_lock(&mtx_socket);
  close(socket_);
  socket_ = 0;
  pthread_mutex_unlock(&mtx_socket);
}


void Connection::send(const string &data) {
  send(data.c_str(), data.length());
}

void Connection::send(const char* buf, size_t len) {
  assert(socket_ > 0);
  pthread_mutex_lock(&mtx_socket);
  write(socket_, buf, len);
  pthread_mutex_unlock(&mtx_socket);
}

void Connection::sendln(const string &data) {
  sendln(data.c_str(), data.length());
}

void Connection::sendln(const char* buf, size_t len) {
  assert(socket_ > 0);

  pthread_mutex_lock(&mtx_socket);
  write(socket_, buf, len);
  write(socket_, "\n", 1);
  pthread_mutex_unlock(&mtx_socket);
}

string Connection::recvln() {
  assert(socket > 0);
  char buf[MAX_LINE_SIZE];
  pthread_mutex_lock(&mtx_socket);
  read(socket_, buf, MAX_LINE_SIZE);
  pthread_mutex_unlock(&mtx_socket);
  return buf;
}
